import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np
import einops
import utils.io as io
from torch.nn import LayerNorm, Linear, MultiheadAttention

class CapInfoNCE_v2(nn.Module,io.WritableToFile):
    def __init__(self,fo=None,fw=None,ku=None,kw=None):
        super().__init__()
        self.fo = fo
        self.fw = fw
        self.ku = ku
        self.kw = kw
        self.temporal_attn = TemporalAttention(input_dim=256, droppath_rate=0.1) # Implementation removed
        
    def forward(self, o, u, w, mask, return_logits=False, temp_attn=None, num_objects=None, topk=None):
        """
        Input:
        :o: BoxToxDo contextualized object features
        :u: BoxToxDu uncontextualized object features
        :w: BwxTwxDw caption word features
        :mask: BwxTw word mask
        """        
        # Compute keys from word features
        Kw = self.kw(w)
        # feat vector size
        D = Kw.size(2)
        Kw = Kw.unsqueeze(1).unsqueeze(3) # Bwx1xTwx1xD
        
        w = w.unsqueeze(1) # Bwx1xTwxDw
        V_w = self.fw(w) # Bwx1xTwxD
        
        mask = mask.unsqueeze(1) # Bwx1xTw
        
        
        # Apply temporal attention along object features
        if temp_attn:
            o_tattn = torch.zeros_like(o)
            for obj_id in range(o.shape[1]):
                o_tattn[:, obj_id, :, :] = self.temporal_attn(o[:, obj_id, :, :].unsqueeze(2)).squeeze()
            o = o_tattn  
            
            u_attn = torch.zeros_like(u)
            for obj_id in range(u.shape[1]):
                u_attn[:, obj_id, :, :] = self.temporal_attn(u[:, obj_id, :, :].unsqueeze(2)).squeeze()
            u = u_attn

        # get frame feats - context and uncontext
        frame_feats_o = torch.unbind(o,2)
        frame_feats_u = torch.unbind(u,2)

        mean_att_frame = []
        total_att = []
        # Compute keys from uncontextualized object features
        for frame_feat_u, frame_feat_o in zip(frame_feats_u, frame_feats_o):
            Ku = self.ku(frame_feat_u)
                        
            # Compute attention using Ku and Kw
            Ku = Ku.unsqueeze(1) # Box1xToxD
            
            # query - vision, key - lang
            att = torch.sum(Kw*Ku,4,keepdim=True) # BwxBoxTwxTox1
            
            # dimensionality of key vectors - D
            att = att / torch.sqrt(torch.tensor(D).float()) # BwxBoxTwxTox1
            att = F.softmax(att,3)
            
            total_att.append(att)
            
            # Compute values from contextualized object features
            frame_feat_o = frame_feat_o.unsqueeze(1) # Box1xToxDo
            V_o = self.fo(frame_feat_o)

            # Compute attended V_o for each word - value * attention score
            att_V_o = torch.sum(att*V_o,3) # BwxBoxTwxDo
        
            # logits = torch.sum(att_V_o*V_w,3) # BwxBoxTw
            mean_att_frame.append(att_V_o)
        
        mean_att_logits = torch.stack(mean_att_frame, 2).mean(2)
        logits = torch.sum(mean_att_logits*V_w,3) # BwxBoxTw
        log_softmax = F.log_softmax(logits,1) # Select image given word
        log_softmax = (1-mask)*log_softmax
        num_non_mask = torch.sum(1-mask,2,keepdim=True) # Bwx1x1
        log_softmax = log_softmax / (num_non_mask + 1e-6)
        # print(log_softmax.size())
        loss = -log_softmax.sum(2).diag().mean()
        # Final attention for image frame
        total_att = torch.stack(total_att,2).mean(2).squeeze(4)
        if return_logits==True:
            return loss, total_att, att_V_o, logits
        
        return loss, total_att, att_V_o
