import os
import glob
import click
import os.path as osp

import utils.io as io
from utils.constants import Constants, ExpConstants
from global_constants import hcstvg_paths, vidstg_paths, hcstvgv2_paths
# from exp.eval_flickr.dataset import  FlickrDatasetConstants
from exp.eval_hcstvg.dataset_videos import  HCSTVGDatasetConstants
from exp.eval_vidstg.dataset_videos import  VidSTGDatasetConstants
from exp.eval_hcstvgv2.dataset_videos import  HCSTVGv2DatasetConstants
from ..models.object_encoder import ObjectEncoderConstants
from ..models.cap_encoder import CapEncoderConstants
from .. import eval_hcstvg_phrase_loc, eval_vidstg_phrase_loc, eval_hcstvgv2_phrase_loc

import utils.io as io

def find_all_model_numbers(model_dir, mode=None):
    model_nums = []
    # Original
    # for name in glob.glob(f'{model_dir}/lang_sup_criterion*'):
    #     model_nums.append(int(name.split('_')[-1]))
    
    for name in glob.glob(f'{model_dir}/object_encoder*'):
        model_nums.append(int(name.split('_')[-1]))
    return sorted(model_nums)


@click.command()
@click.option(
    '--exp_name',
    default='default_exp',
    help='Experiment name')
@click.option(
    '--dataset',
    default='coco',
    type=click.Choice(['hcstvg', 'vidstg', 'hcstvgv2']),
    help='Dataset to use')
@click.option(
    '--no_context',
    is_flag=True,
    help='Apply flag to switch off contextualization')
@click.option(
    '--subset',
    default='val',
    help='subset to run evaluation on')
@click.option(
    '--random_lang',
    is_flag=True,
    help='Apply flag to randomly initialize and train BERT')
@click.option(
    '--cap_info_nce_layers',
    default=2,
    type=int,
    help='Number of layers in lang_sup_criterion')
@click.option(
    '--feat_dim',
    default=256,
    type=int,
    help='Feature dimension')
@click.option(
    '--det_dir',
    default='gdino_feats',
    type=str,
    help='Detection directory')
@click.option(
    '--cfg',
    default='hcstvg_temp.json',
    type=str,
    help='Temporal ground config path')
@click.option(
    '--inf_mode',
    is_flag=True,
    help='Run inference and save results')
@click.option(
    '--temp_attn',
    is_flag=True,
    help='Temporal Attention')
@click.option(
    '--lang',
    is_flag=True,
    help='Language Attention')
@click.option(
    '--clip',
    is_flag=True,
    help='CLIP text encoder')
@click.option(
    '--nl',
    default=1,
    type=int,
    help='Decoder layer number')
@click.option(
    '--nav',
    is_flag=True,
    help='Noun-adj-verb word feats')
@click.option(
    '--phrase',
    is_flag=True,
    help='Include phrases for evaluation')
@click.option(
    '--both',
    is_flag=True,
    help='Evaluate NAV and caption both')
def main(**kwargs):
    if kwargs['dataset']=='hcstvg':
        exp_base_dir = hcstvg_paths['exp_dir']
        data_const = HCSTVGDatasetConstants(kwargs['subset'], kwargs['det_dir'])
    elif kwargs['dataset']=='vidstg':
        exp_base_dir = vidstg_paths['exp_dir']
        data_const = VidSTGDatasetConstants(kwargs['subset'], kwargs['det_dir'])
    elif kwargs['dataset']=='hcstvgv2':
        exp_base_dir = hcstvgv2_paths['exp_dir']
        data_const = HCSTVGv2DatasetConstants(kwargs['subset'], kwargs['det_dir'])
    
    exp_const = ExpConstants(kwargs['exp_name'], exp_base_dir)
    exp_const.model_dir = os.path.join(exp_const.exp_dir,'models')
    exp_const.seed = 42
    exp_const.contextualize = not kwargs['no_context']
    exp_const.random_lang = kwargs['random_lang']
    exp_const.feat_dim = kwargs['feat_dim']
    exp_const.inf_mode = kwargs['inf_mode']
    exp_const.dataset = kwargs['dataset']
    exp_const.temp_attn = kwargs['temp_attn']
    exp_const.lang = kwargs['lang']
    exp_const.clip = kwargs['clip']
    exp_const.nl = kwargs['nl']
    exp_const.nav = kwargs['nav']
    exp_const.phrase = kwargs['phrase']
    exp_const.both = kwargs['both']

    model_const = Constants()
    model_const.object_encoder = ObjectEncoderConstants()
    model_const.object_encoder.context_layer.output_attentions = True
    # if kwargs['dataset']=='flickr':
    #     model_const.object_encoder.object_feature_dim = 2048
    # if kwargs['dataset']=='hcstvg' or kwargs['dataset']=='vidstg':
    model_const.object_encoder.object_feature_dim = exp_const.feat_dim
    model_const.cap_encoder = CapEncoderConstants()
    model_const.cap_encoder.output_attentions = True
    model_const.cap_info_nce_layers = kwargs['cap_info_nce_layers']

    model_nums = find_all_model_numbers(exp_const.model_dir)
    print("Total saved models: ", len(model_nums))
    # , mode=kwargs['sup']
    print(model_nums)
    for num in model_nums:
        # if exp_const.dataset=='vidstg':
        #     if num <= 7000:
        #         continue

        model_const.model_num = num
        # ORIGINAL
        model_const.object_encoder_path = os.path.join(
            exp_const.model_dir,
            f'object_encoder_{model_const.model_num}')
        model_const.lang_sup_criterion_path = os.path.join(
            exp_const.model_dir,
            f'lang_sup_criterion_{model_const.model_num}')
        model_const.multimodal_contrastive_criterion_path = os.path.join(
            exp_const.model_dir,
            f'multimodal_contrast_criterion_{model_const.model_num}')
        
        model_const.temp_ground_path = os.path.join(
            exp_const.model_dir,
            f'temp_model_{model_const.model_num}.pt')
        
        # # BEST AVAILABLE
        # model_const.object_encoder_path = os.path.join(
        #     exp_const.model_dir,
        #     f'best_object_encoder')
        # model_const.lang_sup_criterion_path = os.path.join(
        #     exp_const.model_dir,
        #     f'best_lang_sup_criterion')
        
        if exp_const.random_lang is True:
            model_const.cap_encoder_path = os.path.join(
                exp_const.model_dir,
                f'cap_encoder_{model_const.model_num}')
    
        filename = os.path.join(
            exp_const.exp_dir,
            f'results_{data_const.subset}_{num}.json')
        
        if os.path.exists(filename):
            print(num, io.load_json_object(filename))
            continue
        
        if kwargs['dataset']=='hcstvg':
            eval_hcstvg_phrase_loc.main(exp_const,data_const,model_const,io.load_json_object(kwargs['cfg'])['model']['config'])
        elif kwargs['dataset']=='vidstg':
            eval_vidstg_phrase_loc.main(exp_const,data_const,model_const,io.load_json_object(kwargs['cfg'])['model']['config'])
        elif kwargs['dataset']=='hcstvgv2':
            eval_hcstvgv2_phrase_loc.main(exp_const,data_const,model_const,io.load_json_object(kwargs['cfg'])['model']['config'])

    
    if kwargs['dataset']=='hcstvg' or kwargs['dataset']=='hcstvgv2':
        best_model_num = -1
        best_viou = 0
        best_cls = 0
        best_results = None
        for num in model_nums:
            filename = os.path.join(
                exp_const.exp_dir,
                f'results_{data_const.subset}_{num}.json')
            
            if not os.path.exists(filename):
                continue

            results = io.load_json_object(filename)
            results['model_num'] = str(num)

            if results['viou'] >= best_viou:
                best_results = results
                best_viou = results['viou']
                best_model_num = num

            if results['cls'] >= best_cls:
                best_cls = results['cls']
                
        print('-'*80)
        best_results['model_num'] = best_model_num
        print(best_results)
        filename = os.path.join(
            exp_const.exp_dir,
            f'results_{data_const.subset}_best.json')
        io.dump_json_object(best_results,filename)

        # weights path
        WTS_DIR = osp.join(exp_const.exp_dir, 'models')
        # select top cls and viou and delete rest
        print("Saving best classifier and viou models. Delete rest")
        for num in model_nums:
            filename = os.path.join(
                exp_const.exp_dir,
                f'results_{data_const.subset}_{num}.json')
            if not os.path.exists(filename):
                continue

            results = io.load_json_object(filename)
            
            if results['viou'] < best_viou:
                if results['cls'] < best_cls:
                    print('Deleting:', osp.join(WTS_DIR, 'object_encoder_'+str(num)+'.pth'))
                    os.remove(osp.join(WTS_DIR, 'object_encoder_' +str(num)))
                    os.remove(osp.join(WTS_DIR, 'lang_sup_criterion_'+str(num)))
                    os.remove(osp.join(WTS_DIR, 'temp_model_'+str(num)+'.pt')) 
    else:
        best_model_num_d, best_model_num_i = -1, -1
        best_viou_d, best_viou_i = 0, 0
        best_cls_d, best_cls_i = 0, 0
        best_results_d, best_results_i = None, None
        for num in model_nums:
            filename = os.path.join(
                exp_const.exp_dir,
                f'results_{data_const.subset}_{num}.json')
            
            if not os.path.exists(filename):
                continue

            results = io.load_json_object(filename)
            # declarative
            results[0]['model_num'] = str(num)

            if results[0]['viou'] >= best_viou_d:
                best_results_d = results[0]
                best_viou_d = results[0]['viou']
                best_model_num_d = num

            if results[0]['cls'] >= best_cls_d:
                best_cls_d = results[0]['cls']
            
            # interrogative
            results[1]['model_num'] = str(num)
            
            if results[1]['viou'] >= best_viou_i:
                best_results_i = results[1]
                best_viou_i = results[1]['viou']
                best_model_num_i = num
                
        print('-'*80)
        best_results_d['model_num'] = best_model_num_d
        best_results_i['model_num'] = best_model_num_i
        best_results = [best_results_d, best_results_i]
        print(best_results)
        filename = os.path.join(
            exp_const.exp_dir,
            f'results_{data_const.subset}_best.json')
        io.dump_json_object(best_results,filename)

        # weights path
        WTS_DIR = osp.join(exp_const.exp_dir, 'models')
        # select top cls and viou and delete rest
        print("Saving best classifier and viou models. Delete rest.")
        for num in model_nums:
            filename = os.path.join(
                exp_const.exp_dir,
                f'results_{data_const.subset}_{num}.json')
            if not os.path.exists(filename):
                continue

            results = io.load_json_object(filename)
            if results[0]['viou'] < best_viou_d and results[1]['viou'] < best_viou_i:
                if results[0]['cls'] < best_cls_d:
                    print('Deleting:', osp.join(WTS_DIR, 'object_encoder_'+str(num)+'.pth'))
                    os.remove(osp.join(WTS_DIR, 'object_encoder_' +str(num)))
                    os.remove(osp.join(WTS_DIR, 'lang_sup_criterion_'+str(num)))
                    os.remove(osp.join(WTS_DIR, 'temp_model_'+str(num)+'.pt')) 

if __name__=='__main__':
    main()