import os
import click

from utils.constants import Constants, ExpConstants
from global_constants import hcstvg_paths, hcstvgv2_paths, hcstvg_quant_paths, vidstg_paths
from ..dataset_hcstvg import HCSTVGDatasetConstants
from ..dataset_hcstvgv2 import HCSTVGV2DatasetConstants
from ..dataset_vidstg import VidSTGDatasetConstants
from ..models.object_encoder import ObjectEncoderConstants
from ..models.cap_encoder import CapEncoderConstants
from ..train import main as train
import utils.io as io

import pdb

@click.command()
@click.option(
    '--exp_name',
    default='default_exp',
    help='Experiment name')
@click.option(
    '--dataset',
    default='coco',
    type=click.Choice(['hcstvg', 'vidstg', 'hcstvgv2']),
    help='Dataset to use')
@click.option(
    '--model_num',
    default=-1,
    type=int,
    help='Model number. -1 implies begining of training.')
@click.option(
    '--lr',
    default=1e-5,
    type=float,
    help='Learning rate')
@click.option(
    '--train_batch_size',
    default=256,
    type=int,
    help='Training batch size')
@click.option(
    '--val_batch_size',
    default=256,
    type=int,
    help='Validation batch size')
@click.option(
    '--neg_noun_loss_wt',
    default=1.0,
    type=float,
    help='Weight for negative verb loss')
@click.option(
    '--self_sup_loss_wt',
    default=0.0,
    type=float,
    help='Weight for self supervision loss')
@click.option(
    '--lang_sup_loss_wt',
    default=1.0,
    type=float,
    help='Weight for language supervision loss')
@click.option(
    '--sup_loss_wt',
    default=1.0,
    type=float,
    help='Weight for supervised CE loss')
@click.option(
    '--multimodal_loss_wt',
    default=1.0,
    type=float,
    help='Weight for multimodal contrastive loss')
@click.option(
    '--no_context',
    is_flag=True,
    help='Apply flag to switch off contextualization')
@click.option(
    '--random_lang',
    is_flag=True,
    help='Apply flag to randomly initialize and train BERT')
@click.option(
    '--quant',
    is_flag=True,
    help='Quantifier mode. Not full sentences')
@click.option(
    '--cap_info_nce_layers',
    default=2,
    type=int,
    help='Number of layers in lang_sup_criterion')
@click.option(
    '--feat_dim',
    default=256,
    type=int,
    help='Feature dimension')
@click.option(
    '--det_dir',
    default='gdino_feats',
    type=str,
    help='Detection directory')
@click.option(
    '--cfg',
    default='hcstvg_temp.json',
    type=str,
    help='Temporal ground config path')
@click.option(
    '--epochs',
    default=10,
    type=int,
    help='Training epochs')
@click.option(
    '--log_steps',
    default=50,
    type=int,
    help='Log steps')
@click.option(
    '--val_frequently',
    is_flag=True,
    help='set for val_step=model_save_step else val_step=2*model_save_step')
@click.option(
    '--temp_attn',
    is_flag=True,
    help='Temporal Attention')
@click.option(
    '--clip',
    is_flag=True,
    help='Clip text embeddings')
@click.option(
    '--ii',
    is_flag=True,
    help='Inter-intra')
@click.option(
    '--nav',
    is_flag=True,
    help='Noun-adj-verb word feats')
@click.option(
    '--topk',
    is_flag=True,
    help='Top-k sampling')
@click.option(
    '--both',
    is_flag=True,
    help='Both nav and captions')

def main(**kwargs):
    if kwargs['dataset']=='hcstvg':
        if kwargs['quant'] is True:
            print("Using only quantifiers!!!!")
            exp_base_dir = hcstvg_quant_paths['exp_dir']
        else:
            print("Using full sentences!!!!")
            exp_base_dir = hcstvg_paths['exp_dir']
    elif kwargs['dataset']=='hcstvgv2':
        print("Using full sentences!!!!")
        exp_base_dir = hcstvgv2_paths['exp_dir']
    if kwargs['dataset']=='vidstg':
        print("Using full sentences!!!!")
        exp_base_dir = vidstg_paths['exp_dir']

    exp_const = ExpConstants(kwargs['exp_name'],exp_base_dir)
    exp_const.log_dir = os.path.join(exp_const.exp_dir,'logs')
    exp_const.model_dir = os.path.join(exp_const.exp_dir,'models')
    exp_const.vis_dir = os.path.join(exp_const.exp_dir,'vis')
    exp_const.dataset = kwargs['dataset']
    exp_const.optimizer = 'Adam'
    exp_const.lr = kwargs['lr']
    exp_const.momentum = None
    exp_const.num_epochs = kwargs['epochs']
    exp_const.log_step = kwargs['log_steps']
    # Save models approx. twice every epoch
    # exp_const.model_save_step = 400000//(2*kwargs['train_batch_size']) # 4000=400000/(2*50)
    
    # exp_const.model_save_step = 150000//(2*kwargs['train_batch_size'])
    if exp_const.dataset=='hcstvg':
        exp_const.model_save_step = 10000//(2*kwargs['train_batch_size'])
    elif exp_const.dataset=='hcstvgv2':
        exp_const.model_save_step = 20000//(2*kwargs['train_batch_size'])
    elif exp_const.dataset=='vidstg':
        exp_const.model_save_step = 75000//(2*kwargs['train_batch_size']) # Earlier 60000 because only 1 loop over dataset no 3 loops
    val_freq_factor=2
    if kwargs['val_frequently'] is True:
        val_freq_factor=1
    exp_const.val_step = val_freq_factor*exp_const.model_save_step # set to 1*model_save_step for plotting mi vs perf
    exp_const.num_val_samples = None
    exp_const.train_batch_size = kwargs['train_batch_size']
    exp_const.val_batch_size = kwargs['val_batch_size']
    exp_const.num_workers = 4
    exp_const.seed = 42
    exp_const.neg_noun_loss_wt = kwargs['neg_noun_loss_wt']
    exp_const.self_sup_loss_wt = kwargs['self_sup_loss_wt']
    exp_const.lang_sup_loss_wt = kwargs['lang_sup_loss_wt']
    exp_const.sup_loss_wt = kwargs['sup_loss_wt']
    exp_const.multimodal_loss_wt = kwargs['multimodal_loss_wt']
    exp_const.contextualize = not kwargs['no_context']
    exp_const.random_lang = kwargs['random_lang']
    # exp_const.static = kwargs['static']
    # exp_const.nv_loss = kwargs['nv_loss']
    exp_const.temp_attn = kwargs['temp_attn']
    exp_const.clip = kwargs['clip']
    exp_const.ii = kwargs['ii']
    exp_const.nav = kwargs['nav']
    exp_const.topk = kwargs['topk']
    exp_const.both = kwargs['both']
    
    if exp_const.dataset=='hcstvg':
        DatasetConstants = HCSTVGDatasetConstants
    elif exp_const.dataset=='hcstvgv2':
        DatasetConstants = HCSTVGV2DatasetConstants
    elif exp_const.dataset=='vidstg':
        DatasetConstants = VidSTGDatasetConstants
    
    data_const = {
        'train': DatasetConstants('train', kwargs['det_dir']),
        'val': DatasetConstants('val', kwargs['det_dir']),
    }

    model_const = Constants()
    model_const.model_num = kwargs['model_num']
    
    # Object encoder
    model_const.object_encoder = ObjectEncoderConstants()
    model_const.object_encoder.context_layer.output_attentions = True
    model_const.object_encoder.object_feature_dim = kwargs['feat_dim']
    
    # Caption encoder
    model_const.cap_encoder = CapEncoderConstants()
    model_const.cap_encoder.output_attentions = True
    model_const.cap_info_nce_layers = kwargs['cap_info_nce_layers']
    
    model_const.object_encoder_path = os.path.join(
        exp_const.model_dir,
        f'object_encoder_{model_const.model_num}')
    model_const.self_sup_criterion_path = os.path.join(
        exp_const.model_dir,
        f'self_sup_criterion_{model_const.model_num}')
    model_const.lang_sup_criterion_path = os.path.join(
        exp_const.model_dir,
        f'lang_sup_criterion_{model_const.model_num}')
    model_const.multimodal_contrastive_criterion_path = os.path.join(
        exp_const.model_dir,
        f'multimodal_contrast_criterion_{model_const.model_num}')
    model_const.temp_ground_path = os.path.join(
            exp_const.model_dir,
            f'temp_model_{model_const.model_num}.pt')
    
    print(f'-----------------Training Experiment Details-----------------------')
    print(vars(exp_const))
    print(f'-------------------------------------------------------------------')
    # train(exp_const, data_const, model_const)
    train(exp_const, data_const, model_const, io.load_json_object(kwargs['cfg'])['model']['config'])

if __name__=='__main__':
    main()


'''

'''